@@ -1,7 +1,7 @@
 # BER decoder
-import types
 from pyasn1.type import tag, base, univ, char, useful, tagmap
 from pyasn1.codec.ber import eoo
+from pyasn1.compat.octets import oct2int, octs2ints
 from pyasn1 import error
 
 class AbstractDecoder:
@@ -54,21 +54,21 @@ explicitTagDecoder = ExplicitTagDecoder()
 class IntegerDecoder(AbstractSimpleDecoder):
     protoComponent = univ.Integer(0)
     precomputedValues = {
-        '\x00': 0,
-        '\x01': 1,
-        '\x02': 2,
-        '\x03': 3,
-        '\x04': 4,
-        '\x05': 5,
-        '\x06': 6,
-        '\x07': 7,
-        '\x08': 8,
-        '\x09': 9,
-        'xff': -1,
-        'xfe': -2,
-        'xfd': -3,
-        'xfc': -4,
-        'xfb': -5
+        '\x00':  0,
+        '\x01':  1,
+        '\x02':  2,
+        '\x03':  3,
+        '\x04':  4,
+        '\x05':  5,
+        '\x06':  6,
+        '\x07':  7,
+        '\x08':  8,
+        '\x09':  9,
+        '\xff': -1,
+        '\xfe': -2,
+        '\xfd': -3,
+        '\xfc': -4,
+        '\xfb': -5
         }
     
     def _valueFilter(self, value):
@@ -85,13 +85,13 @@ class IntegerDecoder(AbstractSimpleDecoder):
         if substrate in self.precomputedValues:
             value = self.precomputedValues[substrate]
         else:
-            firstOctet = ord(substrate[0])
+            firstOctet = oct2int(substrate[0])
             if firstOctet & 0x80:
-                value = -1L
+                value = -1
             else:
-                value = 0L
+                value = 0
             for octet in substrate:
-                value = value << 8 | ord(octet)
+                value = value << 8 | oct2int(octet)
             value = self._valueFilter(value)
         return self._createComponent(asn1Spec, tagSet, value), substrate
 
@@ -111,7 +111,7 @@ class BitStringDecoder(AbstractSimpleDecoder):
         if tagSet[0][1] == tag.tagFormatSimple:    # XXX what tag to check?
             if not substrate:
                 raise error.PyAsn1Error('Missing initial octet')
-            trailingBits = ord(substrate[0])
+            trailingBits = oct2int(substrate[0])
             if trailingBits > 7:
                 raise error.PyAsn1Error(
                     'Trailing bits overflow %s' % trailingBits
@@ -122,7 +122,7 @@ class BitStringDecoder(AbstractSimpleDecoder):
                 if p == l:
                     lsb = trailingBits
                 j = 7                    
-                o = ord(substrate[p])
+                o = oct2int(substrate[p])
                 while j >= lsb:
                     b = b + ((o>>j)&0x01,)
                     j = j - 1
@@ -202,14 +202,14 @@ class ObjectIdentifierDecoder(AbstractSimpleDecoder):
             raise error.PyAsn1Error('Empty substrate')
         oid = (); index = 0        
         # Get the first subid
-        subId = ord(substrate[index])
+        subId = oct2int(substrate[index])
         oid = oid + divmod(subId, 40)
 
         index = index + 1
         substrateLen = len(substrate)
         
         while index < substrateLen:
-            subId = ord(substrate[index])
+            subId = oct2int(substrate[index])
             if subId < 128:
                 oid = oid + (subId,)
                 index = index + 1
@@ -220,7 +220,7 @@ class ObjectIdentifierDecoder(AbstractSimpleDecoder):
                 while nextSubId >= 128 and index < substrateLen:
                     subId = (subId << 7) + (nextSubId & 0x7F)
                     index = index + 1
-                    nextSubId = ord(substrate[index])
+                    nextSubId = oct2int(substrate[index])
                 if index == substrateLen:
                     raise error.SubstrateUnderrunError(
                         'Short substrate for OID %s' % oid
@@ -237,9 +237,9 @@ class RealDecoder(AbstractSimpleDecoder):
         substrate = substrate[:length]
         if not length:
             raise error.SubstrateUnderrunError('Short substrate for Real')
-        fo = ord(substrate[0]); substrate = substrate[1:]
+        fo = oct2int(substrate[0]); substrate = substrate[1:]
         if fo & 0x40:  # infinite value
-            value = fo & 0x01 and float('-inf') or float('inf')
+            value = fo & 0x01 and '-inf' or 'inf'
         elif fo & 0x80:  # binary enoding
             if fo & 0x11 == 0:
                 n = 1
@@ -248,19 +248,19 @@ class RealDecoder(AbstractSimpleDecoder):
             elif fo & 0x02:
                 n = 3
             else:
-                n = ord(substrate[0])
+                n = oct2int(substrate[0])
             eo, substrate = substrate[:n], substrate[n:]
             if not eo or not substrate:
                 raise error.PyAsn1Error('Real exponent screwed')
             e = 0
             while eo:         # exponent
                 e <<= 8
-                e |= ord(eo[0])
+                e |= oct2int(eo[0])
                 eo = eo[1:]
             p = 0
             while substrate:  # value
                 p <<= 8
-                p |= ord(substrate[0])
+                p |= oct2int(substrate[0])
                 substrate = substrate[1:]
             if fo & 0x40:    # sign bit
                 p = -p
@@ -268,7 +268,7 @@ class RealDecoder(AbstractSimpleDecoder):
         elif fo & 0xc0 == 0:  # character encoding
             try:
                 if fo & 0x3 == 0x1:  # NR1
-                    value = (long(substrate), 10, 0)
+                    value = (int(substrate), 10, 0)
                 elif fo & 0x3 == 0x2:  # NR2
                     value = float(substrate)
                 elif fo & 0x3 == 0x3:  # NR3
@@ -535,7 +535,7 @@ typeMap = {
 
 ( stDecodeTag, stDecodeLength, stGetValueDecoder, stGetValueDecoderByAsn1Spec,
   stGetValueDecoderByTag, stTryAsExplicitTag, stDecodeValue,
-  stDumpRawValue, stErrorCondition, stStop ) = range(10)
+  stDumpRawValue, stErrorCondition, stStop ) = [x for x in range(10)]
 
 class Decoder:
     defaultErrorState = stErrorCondition
@@ -565,18 +565,18 @@ class Decoder:
                 if firstOctet in self.__tagCache:
                     lastTag = self.__tagCache[firstOctet]
                 else:
-                    t = ord(firstOctet)
+                    t = oct2int(firstOctet)
                     tagClass = t&0xC0
                     tagFormat = t&0x20
                     tagId = t&0x1F
                     if tagId == 0x1F:
-                        tagId = 0L
+                        tagId = 0
                         while 1:
                             if not substrate:
                                 raise error.SubstrateUnderrunError(
                                     'Short octet stream on long tag decoding'
                                     )
-                            t = ord(substrate[0])
+                            t = oct2int(substrate[0])
                             tagId = tagId << 7 | (t&0x7F)
                             substrate = substrate[1:]
                             if not t&0x80:
@@ -604,7 +604,7 @@ class Decoder:
                      raise error.SubstrateUnderrunError(
                          'Short octet stream on length decoding'
                          )
-                firstOctet  = ord(substrate[0])
+                firstOctet  = oct2int(substrate[0])
                 if firstOctet == 128:
                     size = 1
                     length = -1
@@ -623,7 +623,7 @@ class Decoder:
                             (size, len(lengthString), tagSet)
                             )
                     for char in lengthString:
-                        length = (length << 8) | ord(char)
+                        length = (length << 8) | oct2int(char)
                     size = size + 1
                 state = stGetValueDecoder
                 substrate = substrate[size:]
@@ -670,8 +670,7 @@ class Decoder:
                     else:
                         state = stTryAsExplicitTag
             if state == stGetValueDecoderByAsn1Spec:
-                if type(asn1Spec) == types.DictType or \
-                       isinstance(asn1Spec, tagmap.TagMap):
+                if isinstance(asn1Spec, (dict, tagmap.TagMap)):
                     if tagSet in asn1Spec:
                         __chosenSpec = asn1Spec[tagSet]
                     else:
@@ -737,7 +736,7 @@ class Decoder:
                 state = stStop
             if state == stErrorCondition:
                 raise error.PyAsn1Error(
-                    '%s not in asn1Spec: %s' % (tagSet, repr(asn1Spec))
+                    '%r not in asn1Spec: %r' % (tagSet, asn1Spec)
                     )
         return value, substrate
             
