@@ -1,9 +1,8 @@
 # ASN.1 "universal" data types
-import string
-import types
-import operator
+import operator, sys
 from pyasn1.type import base, tag, constraint, namedtype, namedval, tagmap
 from pyasn1.codec.ber import eoo
+from pyasn1.compat import octets
 from pyasn1 import error
 
 # "Simple" ASN.1 types (yet incomplete)
@@ -38,15 +37,25 @@ class Integer(base.AbstractSimpleAsn1Item):
     def __rsub__(self, value): return self.clone(value - self._value)
     def __mul__(self, value): return self.clone(self._value * value)
     def __rmul__(self, value): return self.clone(value * self._value)
-    def __div__(self, value):  return self.clone(self._value / value)
-    def __rdiv__(self, value):  return self.clone(value / self._value)
     def __mod__(self, value): return self.clone(self._value % value)
     def __rmod__(self, value): return self.clone(value % self._value)
     def __pow__(self, value, modulo=None): return self.clone(pow(self._value, value, modulo))
     def __rpow__(self, value): return self.clone(pow(value, self._value))
 
+    if sys.version_info[0] <= 2:
+        def __div__(self, value):  return self.clone(self._value // value)
+        def __rdiv__(self, value):  return self.clone(value // self._value)
+    else:
+        def __truediv__(self, value):  return self.clone(self._value / value)
+        def __rtruediv__(self, value):  return self.clone(value / self._value)
+        def __divmod__(self, value):  return self.clone(self._value // value)
+        def __rdivmod__(self, value):  return self.clone(value // self._value)
+
+        __hash__ = base.AbstractSimpleAsn1Item.__hash__
+
     def __int__(self): return int(self._value)
-    def __long__(self): return long(self._value)
+    if sys.version_info[0] <= 2:
+        def __long__(self): return long(self._value)
     def __float__(self): return float(self._value)    
     def __abs__(self): return abs(self._value)
     def __index__(self): return int(self._value)
@@ -59,27 +68,21 @@ class Integer(base.AbstractSimpleAsn1Item):
     def __ge__(self, value): return self._value >= value
 
     def prettyIn(self, value):
-        if type(value) != types.StringType:
-            return long(value)
+        if not isinstance(value, str):
+            return int(value)
         r = self.__namedValues.getValue(value)
         if r is not None:
             return r
         try:
-            return string.atoi(value)
-        except:
-            try:
-                return string.atol(value)
-            except:
-                raise error.PyAsn1Error(
-                    'Can\'t coerce %s into integer' % value
-                    )
+            return int(value)
+        except ValueError:
+            raise error.PyAsn1Error(
+                'Can\'t coerce %s into integer: %s' % (value, sys.exc_info()[1])
+                )
 
     def prettyOut(self, value):
         r = self.__namedValues.getName(value)
-        if r is not None:
-            return '%s(%s)' % (r, value)
-        else:
-            return str(value)
+        return r is None and str(value) or repr(r)
 
     def getNamedValues(self): return self.__namedValues
 
@@ -184,8 +187,8 @@ class BitString(base.AbstractSimpleAsn1Item):
             self._len = len(self._value)
         return self._len
     def __getitem__(self, i):
-        if type(i) == types.SliceType:
-            return self.clone(operator.getslice(self._value, i.start, i.stop))
+        if isinstance(i, slice):
+            return self.clone(operator.getitem(self._value, i))
         else:
             return self._value[i]
 
@@ -198,7 +201,7 @@ class BitString(base.AbstractSimpleAsn1Item):
         r = []
         if not value:
             return ()
-        elif type(value) == types.StringType:
+        elif isinstance(value, str):
             if value[0] == '\'':
                 if value[-2:] == '\'B':
                     for v in value[1:-2]:
@@ -214,7 +217,7 @@ class BitString(base.AbstractSimpleAsn1Item):
                 elif value[-2:] == '\'H':
                     for v in value[1:-2]:
                         i = 4
-                        v = string.atoi(v, 16)
+                        v = int(v, 16)
                         while i:
                             i = i - 1
                             r.append((v>>i)&0x01)
@@ -224,7 +227,7 @@ class BitString(base.AbstractSimpleAsn1Item):
                         'Bad BIT STRING value notation %s' % value
                         )                
             else:
-                for i in string.split(value, ','):
+                for i in value.split(','):
                     j = self.__namedValues.getValue(i)
                     if j is None:
                         raise error.PyAsn1Error(
@@ -234,7 +237,7 @@ class BitString(base.AbstractSimpleAsn1Item):
                         r.extend([0]*(j-len(r)+1))
                     r[j] = 1
                 return tuple(r)
-        elif type(value) == types.TupleType or type(value) == types.ListType:
+        elif isinstance(value, (tuple, list)):
             r = tuple(value)
             for b in r:
                 if b and b != 1:
@@ -250,63 +253,151 @@ class BitString(base.AbstractSimpleAsn1Item):
                 )
 
     def prettyOut(self, value):
-        return '\"\'%s\'B\"' % string.join(map(str, value), '')
+        return '\"\'%s\'B\"' % ''.join([str(x) for x in value])
 
 class OctetString(base.AbstractSimpleAsn1Item):
     tagSet = baseTagSet = tag.initTagSet(
         tag.Tag(tag.tagClassUniversal, tag.tagFormatSimple, 0x04)
         )
-    def prettyOut(self, value): return repr(value)
-    def prettyIn(self, value):
-        if type(value) == types.StringType:
-            return value
+    defaultBinValue = defaultHexValue = base.noValue
+    encoding = 'us-ascii'
+    def __init__(self, value=None, tagSet=None, subtypeSpec=None,
+                 encoding=None, binValue=None, hexValue=None):
+        if encoding is None:
+            self._encoding = self.encoding
+        else:
+            self._encoding = encoding
+        if binValue is not None:
+            value = self.fromBinaryString(binValue)
+        if hexValue is not None:
+            value = self.fromHexString(hexValue)
+        if value is None or value is base.noValue:
+            value = self.defaultHexValue
+        if value is None or value is base.noValue:
+            value = self.defaultBinValue
+        self.__intValue = None
+        base.AbstractSimpleAsn1Item.__init__(self, value, tagSet, subtypeSpec)
+
+    def clone(self, value=None, tagSet=None, subtypeSpec=None,
+              encoding=None, binValue=None, hexValue=None):
+        if value is None and tagSet is None and subtypeSpec is None and \
+               encoding is None and binValue is None and hexValue is None:
+            return self
+        if value is None and binValue is None and hexValue is None:
+            value = self._value
+        if tagSet is None:
+            tagSet = self._tagSet
+        if subtypeSpec is None:
+            subtypeSpec = self._subtypeSpec
+        if encoding is None:
+            encoding = self._encoding
+        return self.__class__(
+            value, tagSet, subtypeSpec, encoding, binValue, hexValue
+            )
+   
+    if sys.version_info[0] <= 2:
+        def prettyIn(self, value):
+            if isinstance(value, str):
+                return value
+            elif isinstance(value, (tuple, list)):
+                try:
+                    return ''.join([ chr(x) for x in value ])
+                except ValueError:
+                    raise error.PyAsn1Error(
+                        'Bad OctetString initializer \'%s\'' % (value,)
+                        )                
+            else:
+                return str(value)
+    else:
+        def prettyIn(self, value):
+            if isinstance(value, bytes):
+                return value
+            elif isinstance(value, OctetString):
+                return value.asOctets()
+            elif isinstance(value, (tuple, list, map)):
+                try:
+                    return bytes(value)
+                except ValueError:
+                    raise error.PyAsn1Error(
+                        'Bad OctetString initializer \'%s\'' % (value,)
+                        )
+            else:
+                try:
+                    return str(value).encode(self._encoding)
+                except UnicodeEncodeError:
+                    raise error.PyAsn1Error(
+                        'Can\'t encode string \'%s\' with \'%s\' codec' % (value, self._encoding)
+                        )
+                        
+
+    def fromBinaryString(self, value):
+        bitNo = 8; byte = 0; r = ()
+        for v in value:
+            if bitNo:
+                bitNo = bitNo - 1
+            else:
+                bitNo = 7
+                r = r + (byte,)
+                byte = 0
+            if v == '0':
+                v = 0
+            elif v == '1':
+                v = 1
+            else:
+                raise error.PyAsn1Error(
+                    'Non-binary OCTET STRING initializer %s' % (v,)
+                    )
+            byte = byte | (v << bitNo)
+        return octets.ints2octs(r + (byte,))
+        
+    def fromHexString(self, value):            
+        r = p = ()
+        for v in value:
+            if p:
+                r = r + (int(p+v, 16),)
+                p = ()
+            else:
+                p = v
+        if p:
+            r = r + (int(p+'0', 16),)
+        return octets.ints2octs(r)
+
+    def prettyOut(self, value):
+        if sys.version_info[0] <= 2:
+            numbers = tuple([ ord(x) for x in value ])
+        else:
+            numbers = tuple(value)
+        if [ x for x in numbers if x < 32 or x > 126 ]:
+            return '0x' + ''.join([ '%x' % x for x in numbers ])
         else:
             return str(value)
 
-    def purePrettyIn(self, value):
-        if type(value) != types.StringType:
-            return str(value)
-        elif not value:
-            return value
-        elif value[0] == '\'':
-            r = ''
-            if value[-2:] == '\'B':
-                bitNo = 8; byte = 0
-                for v in value[1:-2]:
-                    if bitNo:
-                        bitNo = bitNo - 1
-                    else:
-                        bitNo = 7
-                        r = r + chr(byte)
-                        byte = 0
-                    if v == '0':
-                        v = 0
-                    elif v == '1':
-                        v = 1
-                    else:
-                        raise error.PyAsn1Error(
-                            'Non-binary OCTET STRING initializer %s' % (v,)
-                            )
-                    byte = byte | (v << bitNo)
-                r = r + chr(byte)
-            elif value[-2:] == '\'H':
-                p = ''
-                for v in value[1:-2]:
-                    if p:
-                        r = r + chr(string.atoi(p+v, 16))
-                        p = ''
-                    else:
-                        p = v
-                if p:
-                    r = r + chr(string.atoi(p+'0', 16))
-            else:
-                raise error.PyAsn1Error(
-                    'Bad OCTET STRING value notation %s' % value
-                    )
-            return r
+    def __repr__(self):
+        if self._value is base.noValue:
+            return self.__class__.__name__ + '()'
+        if [ x for x in self.asNumbers() if x < 32 or x > 126 ]:
+            return self.__class__.__name__ + '(hexValue=\'' + ''.join([ '%x' % x for x in self.asNumbers() ])+'\')'
         else:
-            return value
-    
+            return self.__class__.__name__ + '(\'' + self.prettyOut(self._value) + '\')'
+                                
+    if sys.version_info[0] <= 2:
+        def __str__(self): return str(self._value)
+        def __unicode__(self):
+            return self._value.decode(self._encoding, 'ignore')
+        def asOctets(self): return self._value
+        def asNumbers(self):
+            if self.__intValue is None:
+                self.__intValue = tuple([ ord(x) for x in self._value ])
+            return self.__intValue
+    else:
+        def __str__(self): return self._value.decode(self._encoding, 'ignore')
+        def __bytes__(self): return self._value
+        def asOctets(self): return self._value
+        def asNumbers(self):
+            if self.__intValue is None:
+                self.__intValue = tuple(self._value)
+            return self.__intValue
+ 
     # Immutable sequence object protocol
     
     def __len__(self):
@@ -314,23 +405,28 @@ class OctetString(base.AbstractSimpleAsn1Item):
             self._len = len(self._value)
         return self._len
     def __getitem__(self, i):
-        if type(i) == types.SliceType:
-            return self.clone(operator.getslice(self._value, i.start, i.stop))
+        if isinstance(i, slice):
+            return self.clone(operator.getitem(self._value, i))
         else:
             return self._value[i]
 
-    def __add__(self, value): return self.clone(self._value + value)
-    def __radd__(self, value): return self.clone(value + self._value)
+    def __add__(self, value): return self.clone(self._value + self.prettyIn(value))
+    def __radd__(self, value): return self.clone(self.prettyIn(value) + self._value)
     def __mul__(self, value): return self.clone(self._value * value)
     def __rmul__(self, value): return self * value
 
 class Null(OctetString):
-    defaultValue = '' # This is tightly constrained
+    defaultValue = ''.encode()  # This is tightly constrained
     tagSet = baseTagSet = tag.initTagSet(
         tag.Tag(tag.tagClassUniversal, tag.tagFormatSimple, 0x05)
         )
-    subtypeSpec = OctetString.subtypeSpec+constraint.SingleValueConstraint('')
+    subtypeSpec = OctetString.subtypeSpec+constraint.SingleValueConstraint(''.encode())
     
+if sys.version_info[0] <= 2:
+    intTypes = (int, long)
+else:
+    intTypes = int
+
 class ObjectIdentifier(base.AbstractSimpleAsn1Item):
     tagSet = baseTagSet = tag.initTagSet(
         tag.Tag(tag.tagClassUniversal, tag.tagFormatSimple, 0x06)
@@ -347,13 +443,15 @@ class ObjectIdentifier(base.AbstractSimpleAsn1Item):
             self._len = len(self._value)
         return self._len
     def __getitem__(self, i):
-        if type(i) == types.SliceType:
+        if isinstance(i, slice):
             return self.clone(
-                operator.getslice(self._value, i.start, i.stop)
+                operator.getitem(self._value, i)
                 )
         else:
             return self._value[i]
 
+    def __str__(self): return self.prettyPrint()
+    
     def index(self, suboid): return self._value.index(suboid)
 
     def isPrefixOf(self, value):
@@ -366,50 +464,50 @@ class ObjectIdentifier(base.AbstractSimpleAsn1Item):
 
     def prettyIn(self, value):
         """Dotted -> tuple of numerics OID converter"""
-        if type(value) is types.TupleType:
+        if isinstance(value, tuple):
             pass
         elif isinstance(value, ObjectIdentifier):
             return tuple(value)        
-        elif type(value) is types.StringType:
+        elif isinstance(value, str):
             r = []
-            for element in filter(None, string.split(value, '.')):
+            for element in [ x for x in value.split('.') if x != '' ]:
                 try:
-                    r.append(string.atoi(element, 0))
-                except string.atoi_error:
-                    try:
-                        r.append(string.atol(element, 0))
-                    except string.atol_error, why:                        
-                        raise error.PyAsn1Error(
-                            'Malformed Object ID %s at %s: %s' %
-                            (str(value), self.__class__.__name__, why)
-                            )
+                    r.append(int(element, 0))
+                except ValueError:
+                    raise error.PyAsn1Error(
+                        'Malformed Object ID %s at %s: %s' %
+                        (str(value), self.__class__.__name__, sys.exc_info()[1])
+                        )
             value = tuple(r)
-
-            pass
         else:
-            value = tuple(value)
+            try:
+                value = tuple(value)
+            except TypeError:
+                raise error.PyAsn1Error(
+                        'Malformed Object ID %s at %s: %s' %
+                        (str(value), self.__class__.__name__,sys.exc_info()[1])
+                        )
 
         for x in value:
-            if x < 0:
+            if not isinstance(x, intTypes) or x < 0:
                 raise error.PyAsn1Error(
-                    'Negative sub-ID in %s at %s' % (value, self.__class__.__name__)
+                    'Invalid sub-ID in %s at %s' % (value, self.__class__.__name__)
                     )
     
         return value
 
-    def prettyOut(self, value):
-        """Tuple of numerics -> dotted string OID converter"""
-        r = []
-        for subOid in value:
-            r.append(str(subOid))
-            if r[-1] and r[-1][-1] == 'L':
-                r[-1][-1] = r[-1][:-1]
-        return string.join(r, '.')
-
+    def prettyOut(self, value): return '.'.join([ str(x) for x in value ])
+    
 class Real(base.AbstractSimpleAsn1Item):
-    _plusInf = float('inf')
-    _minusInf = float('-inf')
-    _inf = (_plusInf, _minusInf)
+    try:
+        _plusInf = float('inf')
+        _minusInf = float('-inf')
+        _inf = (_plusInf, _minusInf)
+    except ValueError:
+        # Infinity support is platform and Python dependent
+        _plusInf = _minusInf = None
+        _inf = ()
+
     tagSet = baseTagSet = tag.initTagSet(
         tag.Tag(tag.tagClassUniversal, tag.tagFormatSimple, 0x09)
         )
@@ -422,9 +520,9 @@ class Real(base.AbstractSimpleAsn1Item):
         return m, b, e
 
     def prettyIn(self, value):
-        if type(value) == types.TupleType and len(value) == 3:
+        if isinstance(value, tuple) and len(value) == 3:
             for d in value:
-                if type(d) not in (types.IntType, types.LongType):
+                if not isinstance(d, intTypes):
                     raise error.PyAsn1Error(
                         'Lame Real value syntax: %s' % (value,)
                         )
@@ -435,28 +533,24 @@ class Real(base.AbstractSimpleAsn1Item):
             if value[1] == 10:
                 value = self.__normalizeBase10(value)
             return value
-        elif type(value) in (types.IntType, types.LongType):
+        elif isinstance(value, intTypes):
             return self.__normalizeBase10((value, 10, 0))
-        elif type(value) == types.FloatType:
-            if value in self._inf:
+        elif isinstance(value, float):
+            if self._inf and value in self._inf:
                 return value
             else:
                 e = 0
-                while long(value) != value:
+                while int(value) != value:
                     value = value * 10
                     e = e - 1
-                return self.__normalizeBase10((long(value), 10, e))
-        
+                return self.__normalizeBase10((int(value), 10, e))
         elif isinstance(value, Real):
             return tuple(value)
-        elif type(value) == types.StringType:  # handle infinite literal
+        elif isinstance(value, str):  # handle infinite literal
             try:
-                inf = float(value)
+                return float(value)
             except ValueError:
                 pass
-            else:
-                if inf in self._inf:
-                    return inf
         raise error.PyAsn1Error(
             'Bad real value syntax: %s' % (value,)
             )
@@ -479,15 +573,23 @@ class Real(base.AbstractSimpleAsn1Item):
     def __rmul__(self, value): return self * value
     def __sub__(self, value): return self.clone(float(self) - value)
     def __rsub__(self, value): return self.clone(value - float(self))
-    def __div__(self, value): return self.clone(float(self) / value)
-    def __rdiv__(self, value): return self.clone(value / float(self))
     def __mod__(self, value): return self.clone(float(self) % value)
     def __rmod__(self, value): return self.clone(value % float(self))
     def __pow__(self, value, modulo=None): return self.clone(pow(float(self), value, modulo))
     def __rpow__(self, value): return self.clone(pow(value, float(self)))
 
+    if sys.version_info[0] <= 2:
+        def __div__(self, value): return self.clone(float(self) / value)
+        def __rdiv__(self, value): return self.clone(value / float(self))
+    else:
+        def __truediv__(self, value): return self.clone(float(self) / value)
+        def __rtruediv__(self, value): return self.clone(value / float(self))
+        def __divmod__(self, value): return self.clone(float(self) // value)
+        def __rdivmod__(self, value): return self.clone(value // float(self))
+
     def __int__(self): return int(float(self))
-    def __long__(self): return long(float(self))
+    if sys.version_info[0] <= 2:
+        def __long__(self): return long(float(self))
     def __float__(self):
         if self._value in self._inf:
             return self._value
@@ -504,7 +606,11 @@ class Real(base.AbstractSimpleAsn1Item):
     def __gt__(self, value): return float(self) > value
     def __ge__(self, value): return float(self) >= value
 
-    def __nonzero__(self): return float(self) and 1 or 0
+    if sys.version_info[0] <= 2:
+        def __nonzero__(self): return bool(float(self))
+    else:
+        def __bool__(self): return bool(float(self))
+        __hash__ = base.AbstractSimpleAsn1Item.__hash__
 
     def __getitem__(self, idx):
         if self._value in self._inf:
@@ -556,7 +662,7 @@ class SetOf(base.AbstractConstructedAsn1Item):
                 self._componentValues[idx] = self._componentType.clone()
                 self._componentValuesSet = self._componentValuesSet + 1
             return self
-        elif type(value) != types.InstanceType:
+        elif not isinstance(value, base.Asn1Item):
             if self._componentType is None:
                 raise error.PyAsn1Error('Component type not defined')
             if isinstance(self._componentType, base.AbstractSimpleAsn1Item):
@@ -606,13 +712,13 @@ class SequenceAndSetBase(base.AbstractConstructedAsn1Item):
             self._componentTypeLen = len(self._componentType)
 
     def __getitem__(self, idx):
-        if type(idx) == types.StringType:
+        if isinstance(idx, str):
             return self.getComponentByName(idx)
         else:
             return base.AbstractConstructedAsn1Item.__getitem__(self, idx)
 
     def __setitem__(self, idx, value):
-        if type(idx) == types.StringType:
+        if isinstance(idx, str):
             self.setComponentByName(idx, value)
         else:
             base.AbstractConstructedAsn1Item.__setitem__(self, idx, value)
@@ -637,8 +743,7 @@ class SequenceAndSetBase(base.AbstractConstructedAsn1Item):
                 )
         t = self._componentType[idx].getType()
         if not t.isSuperTypeOf(value):
-            raise error.PyAsn1Error('Component type error %s vs %s' %
-                                    (repr(t), repr(value)))
+            raise error.PyAsn1Error('Component type error %r vs %r' % (t, value))
 
     def getComponentByName(self, name):
         return self.getComponentByPosition(
@@ -666,7 +771,7 @@ class SequenceAndSetBase(base.AbstractConstructedAsn1Item):
                 self._componentValues[idx] = self._componentType.getTypeByPosition(idx).clone()
                 self._componentValuesSet = self._componentValuesSet + 1
             return self
-        elif type(value) != types.InstanceType:
+        elif not isinstance(value, base.Asn1Item):
             t = self._componentType.getTypeByPosition(idx)
             if isinstance(t, base.AbstractSimpleAsn1Item):
                 value = t.clone(value=value)
@@ -705,7 +810,7 @@ class SequenceAndSetBase(base.AbstractConstructedAsn1Item):
             elif not self._componentType[idx].isOptional:
                 if self.getComponentByPosition(idx) is None:
                     raise error.PyAsn1Error(
-                        'Uninitialized component #%s at %s' % (idx, repr(self))
+                        'Uninitialized component #%s at %r' % (idx, self)
                         )
 
     def prettyPrint(self, scope=0):
@@ -794,10 +899,34 @@ class Choice(Set):
     typeId = 5
     _currentIdx = None
 
-    def __cmp__(self, other):
+    def __eq__(self, other):
         if self._componentValues:
-            return cmp(self._componentValues[self._currentIdx], other)
-        return -1
+            return self._componentValues[self._currentIdx] == other
+        return NotImplemented
+    def __ne__(self, other):
+        if self._componentValues:
+            return self._componentValues[self._currentIdx] != other
+        return NotImplemented
+    def __lt__(self, other):
+        if self._componentValues:
+            return self._componentValues[self._currentIdx] < other
+        return NotImplemented
+    def __le__(self, other):
+        if self._componentValues:
+            return self._componentValues[self._currentIdx] <= other
+        return NotImplemented
+    def __gt__(self, other):
+        if self._componentValues:
+            return self._componentValues[self._currentIdx] > other
+        return NotImplemented
+    def __ge__(self, other):
+        if self._componentValues:
+            return self._componentValues[self._currentIdx] >= other
+        return NotImplemented
+    if sys.version_info[0] <= 2:
+        def __nonzero__(self, other): return bool(self._componentValues)
+    else:
+        def __bool__(self, other): return bool(self._componentValues)
 
     def __len__(self): return self._currentIdx is not None and 1 or 0
     
@@ -836,7 +965,7 @@ class Choice(Set):
                 self._componentValuesSet = 1
                 self._currentIdx = idx
             return self
-        elif type(value) != types.InstanceType:
+        elif not isinstance(value, base.Asn1Item):
             value = self._componentType.getTypeByPosition(idx).clone(
                 value=value
                 )
